#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/skbuff.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/tcp.h>
#include <linux/types.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/tcp.h>

MODULE_DESCRIPTION("JetMax Host");
MODULE_AUTHOR("saujain (saujain@tamu.edu)");
MODULE_LICENSE("GPL");

#define MAX(x,y) (x > y ? x : y)
#define MIN(x,y) (x < y ? x : y)
#define ABS(x) (x > 0 ? x : (-1)*(x))

#define SEQ_LT(a,b)		((int)((a)-(b)) < 0)
#define SEQ_LEQ(a,b)	((int)((a)-(b)) <= 0)
#define SEQ_GT(a,b)		((int)((a)-(b)) > 0)
#define SEQ_GEQ(a,b)	((int)((a)-(b)) >= 0)

struct jetmaxhdr {
	u8 A:1;         // Approve or disapprove
	u8 I:1;         // Interval
	u8 A_F:1;       // Approve or disapprove feedback
	u8 I_F:1;       // Interval feedback
	u8 U:4;         // Unused
	u8 RT;
	u8 RC;
	u8 RS;
	float pkt_loss;
	float fair_rate;
	float proposed_pktsize;
	float ipi;
	u8 RS_feedback;
	float pkt_loss_feedback;
	float fair_rate_feedback;
};

void display_pkt(struct jetmaxhdr *xhdr) {
	if(xhdr != NULL) {
		printk(KERN_INFO "%d,%d,%d,%d,%d,%d,%d,%ld,%d,%ld,%d,%d,%ld,%d\n",
		xhdr->A,xhdr->I,xhdr->A_F,xhdr->I_F,xhdr->RT,xhdr->RC,xhdr->RS,
		(long) xhdr->pkt_loss*1000,(int)xhdr->fair_rate,
		(long)xhdr->proposed_pktsize,(int)xhdr->ipi*1000,xhdr->RS_feedback,
		(unsigned long)xhdr->pkt_loss_feedback*1000,(int)xhdr->fair_rate_feedback);
	}
}	

struct perflow {
	int hash;
	struct tcp_sock *to;
	unsigned long bw_data;
	unsigned long bw_last_update;
	struct timer_list bw_timer;
	int is_bw_timer_up;
	int zero_data_count;
	unsigned int ack_seqno;
	float fair_rate;
	rwlock_t lk;
};

static int PROTOCOL = 202;
static int TAU = 60;
static int TAU_SCALE = 100;
static int IS_RX = 0;
static double tau = 0.6;
static struct net_protocol jetmaxproto;
static struct perflow flows[1024];
int (*curr_queue_xmit) (struct sk_buff* skb, int ipfragok);
int netheader_len = 0;
static int BW_INTERVAL = 100;
static int KILL_TIMER_CNT = 10;
void bw_timer_handler(unsigned long);

module_param(PROTOCOL,int,S_IRUGO);
module_param(TAU,int,S_IRUGO);
module_param(TAU_SCALE,int,S_IRUGO);
module_param(BW_INTERVAL,int,S_IRUGO);
module_param(IS_RX,int,S_IRUGO);
module_param(KILL_TIMER_CNT,int,S_IRUGO);

extern struct net_protocol tcp_protocol;
extern struct tcp_func ipv4_specific;
extern void tcp_jetmax_output(unsigned long data);

struct perflow* get_flow(uint32_t daddr, uint32_t saddr, uint16_t dport, uint16_t sport) 
{
  unsigned char c = 0;
  unsigned int h = (saddr ^ sport) ^ (daddr ^ dport);
  h ^= h >> 16;
  h ^= h >> 8;
  c = (unsigned char)h;
  flows[c].hash = c;
  return &(flows[c]);
}

//IP->JetMax->TCP
//Possible Cases: 1) Receiver: receives a Data packet 
//                2) Sender: receives an Ack packet
//                3) Sender: receives a Data-Ack packet
int jetmax_pkt_handler(struct sk_buff* skb) 
{
	struct perflow* pd = 0;
	struct jetmaxhdr* hdr = (struct jetmaxhdr*)skb->data;
	int cpu;
	u8 RS;
	u8 I;
	u8 A;
	float pkt_loss;
	float fair_rate;
	unsigned int ack_sequence_no;
	double diff, tmp_rate_proposed_, temp_;

	u8 A_F;
	u8 I_F;
	u8 RS_feedback;
	float pkt_loss_feedback;
	float fair_rate_feedback;

	//Record variables.
	RS = hdr->RS;
	I = hdr->I;
	A = hdr->A;
	pkt_loss = hdr->pkt_loss;
	fair_rate = hdr->fair_rate;
	A_F = hdr->A_F;
	I_F = hdr->I_F;
	RS_feedback = hdr->RS_feedback;
	pkt_loss_feedback = hdr->pkt_loss_feedback;
	fair_rate_feedback = hdr->fair_rate_feedback;

	//Remove Jetmax header
	skb_pull(skb,sizeof(struct jetmaxhdr));
	skb->h.th = (struct tcphdr*)skb->data;
	
	pd = get_flow(skb->nh.iph->daddr,skb->nh.iph->saddr,
		ntohs((short)(skb->h.th->dest)),
		ntohs((short)(skb->h.th->source)));

	ack_sequence_no = ntohl(skb->h.th->ack_seq);

	if(pd == NULL) {
		printk(KERN_CRIT "#jetmax_pkt_handler: Null value of pd obtained.\n");
		return 0;
	}
		
	if(pd->to!=NULL) {
		cpu = get_cpu();
		pd->to->jetmax.netheader_len = netheader_len;
		pd->to->jetmax.A_F = A;
		pd->to->jetmax.I_F = I;
		pd->to->jetmax.RS_feedback = RS;
		pd->to->jetmax.pkt_loss_feedback = pkt_loss;
		pd->to->jetmax.fair_rate_feedback = fair_rate;
		pd->ack_seqno = ack_sequence_no;
		pd->fair_rate = fair_rate_feedback;
		
		if (pd->to->jetmax.active_ == 1)
		{
			if (SEQ_GT(ack_sequence_no, pd->to->jetmax.max_pkt_seqno_) && 
			I_F != pd->to->jetmax.I && pkt_loss_feedback != 2)
			{
				pd->to->jetmax.cur_loss_seqno_ ++;
				pd->to->jetmax.I = I_F;
				pd->to->jetmax.max_pkt_seqno_ = ack_sequence_no;
			}
			if (RS_feedback != pd->to->jetmax.RT && RS_feedback > 0 && 
				pd->to->jetmax.switch_in_progress_ == 0 && 
				pd->to->jetmax.test_in_progress_ == 0) {
				if (pd->to->jetmax.RT == 0)
				{
					pd->to->jetmax.RT = RS_feedback;
					pd->to->jetmax.switch_in_progress_ = 1;
					pd->to->jetmax.seqno_last_rtt = pd->to->snd_nxt;
					pd->to->jetmax.is_rtt_expired_ = 0;
				}
				else
				{
					diff = ABS(pkt_loss_feedback - pd->to->jetmax.pkt_loss_);
					if (diff >= 0.1)
					{
						if (pd->to->jetmax.test_switch_ == 0)
						{
							pd->to->jetmax.test_in_progress_ = 1;
							pd->to->jetmax.test_switch_ = 1;
							pd->to->jetmax.interval_cnt_ = 3;
						}
						else
						{
							pd->to->jetmax.RT = RS_feedback;
							pd->to->jetmax.test_switch_ = 0;
							pd->to->jetmax.switch_in_progress_ = 1;
							pd->to->jetmax.seqno_last_rtt = pd->to->snd_nxt;
							pd->to->jetmax.is_rtt_expired_ = 0;
						}
					}
					else if (pd->to->jetmax.rate_ > fair_rate_feedback && 
					fair_rate_feedback != -1 && fair_rate_feedback != 0)
					{
						if (pd->to->jetmax.stationary != 0)
						{
							if (pd->to->jetmax.test_switch_ == 0)
							{
								pd->to->jetmax.test_in_progress_ = 1;
								pd->to->jetmax.test_switch_ = 1;
								pd->to->jetmax.interval_cnt_ = 3;				
							}
							else
							{
								pd->to->jetmax.RT = RS_feedback;
								pd->to->jetmax.test_switch_ = 0;
								pd->to->jetmax.switch_in_progress_ = 1;
								pd->to->jetmax.seqno_last_rtt = pd->to->snd_nxt;
								pd->to->jetmax.is_rtt_expired_ = 0;
							}
						}
					}
				}
			}	
			else if (RS_feedback == pd->to->jetmax.RT) {
					pd->to->jetmax.test_switch_ = 0;
			}

			if (pd->to->jetmax.switch_in_progress_ ==1 && 
					SEQ_GT(ack_sequence_no , pd->to->jetmax.seqno_last_rtt))
			{
				pd->to->jetmax.is_rtt_expired_ = 1;
				pd->to->jetmax.max_loss_seqno_ = pd->to->jetmax.cur_loss_seqno_;
				pd->to->jetmax.switch_in_progress_ = 2;
			}

			if(pkt_loss_feedback != 2 && RS_feedback > 0)
			{	
				if (ack_sequence_no > pd->to->jetmax.seqno_last_rtt 
						&& pd->to->jetmax.is_rtt_expired_ == 0)
					pd->to->jetmax.is_rtt_expired_ = 1;

				if (pd->to->jetmax.cur_loss_seqno_ >= pd->to->jetmax.max_loss_seqno_ &&
						pd->to->jetmax.is_rtt_expired_ == 0)
				{
					pd->to->jetmax.max_loss_seqno_ = pd->to->jetmax.cur_loss_seqno_;
					pd->to->jetmax.max_pkt_seqno_ = ack_sequence_no;
				}

				if(pd->to->jetmax.cur_loss_seqno_ > pd->to->jetmax.max_loss_seqno_ && 
						pd->to->jetmax.is_rtt_expired_ == 1)
				{
					pd->to->jetmax.max_loss_seqno_ = pd->to->jetmax.cur_loss_seqno_;

					if (pd->to->jetmax.switch_in_progress_ == 2)
						pd->to->jetmax.switch_in_progress_ = 3;

					if (pd->to->jetmax.test_in_progress_ != 0)
						pd->to->jetmax.test_in_progress_ ++;
					
					if (pd->to->jetmax.switch_in_progress_ == 0 && 
					pd->to->jetmax.test_in_progress_ == 0)
					{
						pd->to->jetmax.test_switch_ = 0;
						pd->to->jetmax.pkt_loss_ = pkt_loss_feedback;
						pd->to->jetmax.rate_avg_ = fair_rate_feedback;
						pd->to->jetmax.max_pkt_seqno_ = ack_sequence_no;
						pd->to->jetmax.current_ACK_pkt_seqno_ = ack_sequence_no;
						pd->to->jetmax.ref_rate_ = pd->to->jetmax.rate_proposed_;
						pd->to->jetmax.is_confirmed_++;
						if (pd->to->jetmax.is_confirmed_ >= 2)
						{
							if (A_F != 0)
							{
								temp_ = pd->to->jetmax.rate_proposed_;
								pd->to->jetmax.rate_proposed_ = pd->to->jetmax.ref_rate_ - 
											tau*(pd->to->jetmax.ref_rate_ - pd->to->jetmax.rate_avg_);

								if (pd->to->jetmax.rate_proposed_ <= temp_)	
									temp_ = pd->to->jetmax.rate_proposed_;

								if(temp_ != pd->to->jetmax.rate_) {

									pd->to->jetmax.rate_ = temp_;
									pd->to->jetmax.bytes_per_send_ = pd->to->jetmax.rate_;

									del_timer_sync(&(pd->to->jetmax.send_timer_));
									init_timer(&(pd->to->jetmax.send_timer_));
									pd->to->jetmax.send_timer_.expires = (jiffies+1);
									pd->to->jetmax.send_timer_.data = (unsigned long)(pd->to);
									pd->to->jetmax.send_timer_.function = tcp_jetmax_output;
									add_timer(&(pd->to->jetmax.send_timer_));
									pd->to->jetmax.istimer_up_ = 1;
		
								}

								pd->to->jetmax.psize_proposed_ = pd->to->jetmax.psize_ * (double) pd->to->jetmax.rate_proposed_ / (double) pd->to->jetmax.rate_;

								pd->to->jetmax.is_confirmed_ = 0;
								pd->to->jetmax.seqno_last_rtt = pd->to->jetmax.cur_pkt_seqno_;
								pd->to->jetmax.is_rtt_expired_ = 0;
								/*
								 printk(KERN_INFO "Rate approved:%ld,%ld,%ld,%d,%d\n",
											jiffies,(unsigned long)pd->to->jetmax.rate_,
											(unsigned long)fair_rate_feedback,RS_feedback,
											(int)(1000*pd->to->jetmax.pkt_loss_));
								*/
							}
							else
							{
								tmp_rate_proposed_ = pd->to->jetmax.ref_rate_ - tau * 
								(pd->to->jetmax.ref_rate_ - pd->to->jetmax.rate_avg_);
								if (tmp_rate_proposed_ <= pd->to->jetmax.rate_)
								{
									pd->to->jetmax.rate_proposed_ = tmp_rate_proposed_;

									if(pd->to->jetmax.rate_ != pd->to->jetmax.rate_proposed_) {

										pd->to->jetmax.rate_ = pd->to->jetmax.rate_proposed_;
										pd->to->jetmax.bytes_per_send_ = pd->to->jetmax.rate_;

										del_timer_sync(&(pd->to->jetmax.send_timer_));
										init_timer(&(pd->to->jetmax.send_timer_));
										pd->to->jetmax.send_timer_.expires = (jiffies+1);
										pd->to->jetmax.send_timer_.data = (unsigned long)(pd->to);
										pd->to->jetmax.send_timer_.function = tcp_jetmax_output;
										add_timer(&(pd->to->jetmax.send_timer_));
										pd->to->jetmax.istimer_up_ = 1;
					
									}

									pd->to->jetmax.psize_proposed_ = pd->to->jetmax.psize_ * (double) pd->to->jetmax.rate_proposed_ / (double)pd->to->jetmax.rate_;
									pd->to->jetmax.is_confirmed_ = 0;
									pd->to->jetmax.seqno_last_rtt = pd->to->jetmax.cur_pkt_seqno_;
									pd->to->jetmax.is_rtt_expired_ = 0;
									/*
									printk(KERN_INFO "Rate not approved:%ld,%ld,%ld,%d,%d\n",
										jiffies,(unsigned long)pd->to->jetmax.rate_,
										(unsigned long)fair_rate_feedback,RS_feedback,
										(int)(1000*pd->to->jetmax.pkt_loss_));
									*/
								}
								else if (tmp_rate_proposed_ < pd->to->jetmax.rate_proposed_)
								{
									pd->to->jetmax.rate_proposed_ = tmp_rate_proposed_;
									pd->to->jetmax.psize_proposed_ = pd->to->jetmax.psize_ * (double) pd->to->jetmax.rate_proposed_ / (double) pd->to->jetmax.rate_;
									pd->to->jetmax.is_confirmed_ = 0;
									pd->to->jetmax.seqno_last_rtt = pd->to->jetmax.cur_pkt_seqno_;
									pd->to->jetmax.is_rtt_expired_ = 0;
								}
							}
						}

						diff = ABS(pd->to->jetmax.rate_ - fair_rate_feedback);
						if (diff < fair_rate_feedback * 0.1)
							pd->to->jetmax.stationary = 1;
						else
							pd->to->jetmax.stationary = 0;	
					}
					if (pd->to->jetmax.test_in_progress_ == pd->to->jetmax.interval_cnt_)
						pd->to->jetmax.test_in_progress_ = 0;

					if (pd->to->jetmax.switch_in_progress_ == 3)
						pd->to->jetmax.switch_in_progress_ = 0;
				}
			}
		}
		put_cpu();
	}
	
	return tcp_protocol.handler(skb);
}


//TCP->JetMax->IP
//Possible cases: 1) Sender: Sending a data packet 
//		  2) Receiver: Sending an Ack 
//		  3) Receiver: Sending a Data-Ack packet
int jetmax_to_ip_pkt_dispatcher(struct sk_buff* skb, int ipfragok) 
{
	int cpu = 0;
	int len = 0;
	struct jetmaxhdr* hdr = 0;
	struct perflow* pd = 0;
	struct inet_sock* io = inet_sk(skb->sk);
	struct tcphdr* thdr = (struct tcphdr*)skb->h.th;
	struct tcp_skb_cb *tcb = TCP_SKB_CB(skb);
	len = skb->len; //Will return a length of 1448

	pd = get_flow(io->daddr,io->saddr,ntohs((short)(thdr->dest)),
		ntohs((short)(thdr->source)));

	if(pd==NULL) {
		printk(KERN_CRIT "#Host-dispatcher: Null value of pd obtained.\n");
		return 0;
	}

	if((tcb->flags & TCPCB_FLAG_SYN))
	{
		pd->to = NULL;
	}

	pd->to = tcp_sk(skb->sk);
	pd->to->rcv_snd = 107374182;
	pd->to->rcv_ssthresh = 107374182;
	hdr = (struct jetmaxhdr*)skb_push(skb,sizeof(struct jetmaxhdr));
	hdr->I = 0;
	hdr->A = 1;
	hdr->A_F = 0;
	hdr->I_F = 0;
	hdr->RC = 0;
	hdr->RS = 0;
	hdr->RT = 0;
	hdr->pkt_loss = 2;
	hdr->fair_rate = -1;
	hdr->proposed_pktsize = 0;
	hdr->ipi = 0;
	hdr->RS_feedback = 0;
	hdr->pkt_loss_feedback = 0;
	hdr->fair_rate_feedback = 0;
	
	skb->sk->sk_protocol = PROTOCOL;

	if(pd!=NULL) {
		cpu = get_cpu();
		hdr->I = pd->to->jetmax.I;
		hdr->I_F = pd->to->jetmax.I_F;
		hdr->A_F = pd->to->jetmax.A_F;
		hdr->RT = pd->to->jetmax.RT;

		if(pd->to->jetmax.psize_proposed_ == 0)
			hdr->proposed_pktsize = pd->to->jetmax.psize_;
		else
			hdr->proposed_pktsize = pd->to->jetmax.psize_proposed_;

		hdr->ipi = pd->to->jetmax.ipi;

		hdr->RS_feedback = pd->to->jetmax.RS_feedback;
		hdr->pkt_loss_feedback = pd->to->jetmax.pkt_loss_feedback;
		hdr->fair_rate_feedback = pd->to->jetmax.fair_rate_feedback;
		put_cpu();
		
		if(pd->to->jetmax.active_ == 1 && IS_RX == 0) {
			if(pd->is_bw_timer_up == 0 && pd->zero_data_count <= KILL_TIMER_CNT)  {
				pd->bw_last_update = jiffies;
				pd->bw_data = (len + sizeof(struct jetmaxhdr) + 20);
				init_timer(&(pd->bw_timer));
				pd->bw_timer.data = (unsigned long)pd;
				pd->bw_timer.function = bw_timer_handler;
				pd->bw_timer.expires = jiffies + (BW_INTERVAL*HZ)/1000;
				add_timer(&(pd->bw_timer));
				pd->is_bw_timer_up = 1;
			} else {
				write_lock(&(pd->lk));
				pd->bw_data += (len + sizeof(struct jetmaxhdr) + 20);
				write_unlock(&(pd->lk));
			}
		}
	}
	//printk(KERN_INFO "%d\n",1);
	//display_pkt(hdr);
	return curr_queue_xmit(skb,ipfragok);
}

void bw_timer_handler(unsigned long pd) {
	int interval = 0;
	unsigned long traffic = 0;
	struct perflow *str = (struct perflow*)pd;
	del_timer_sync(&(str->bw_timer));
	write_lock(&(str->lk));
	traffic = str->bw_data;
	if(traffic == 0)
		str->zero_data_count++;
	else
		str->zero_data_count = 0;
	str->bw_data = 0;
	interval = jiffies - str->bw_last_update;
	str->bw_last_update = jiffies;
	write_unlock(&(str->lk));

	if(interval > 0) {
		printk(KERN_INFO "%d:%ld:%ld:%d:%d:%d:%d:%d:%d\n",
											str->hash,
											str->bw_last_update,
											traffic/interval,(int)str->to->jetmax.rate_,
											(int)str->to->jetmax.rate_avg_,interval,
											str->to->snd_cwnd,str->to->srtt/8,str->to->jetmax.RT
											//(int)(1000*str->to->jetmax.pkt_loss_),
											//(int)str->fair_rate,
											//str->to->jetmax.max_pkt_seqno_,
											//str->ack_seqno,str->to->jetmax.cur_loss_seqno_
											);
	} else {
		printk(KERN_CRIT "#bw_timer_handler: interval is <= 0");
	}

	str->bw_data = 0;
	if(str->to->jetmax.active_ == 1 && IS_RX == 0 && 
						str->zero_data_count <= KILL_TIMER_CNT) {
		init_timer(&(str->bw_timer));
		str->bw_timer.data = (unsigned long)pd;
		str->bw_timer.function = bw_timer_handler;
		str->bw_timer.expires = str->bw_last_update + (BW_INTERVAL*HZ)/1000;
		add_timer(&(str->bw_timer));
	}
}

void jetmax_err (struct sk_buff* skb, u32 info) 
{
	//printk("#JetMax error handler called : %u\n",info);
	tcp_protocol.err_handler(skb,info);
}


static int jetmax_host_init_module(void)
{
	int i = 0;
	struct tcp_func* tf = &ipv4_specific;
	tau = (double)TAU/(double)TAU_SCALE;
	jetmaxproto.handler = jetmax_pkt_handler;
	jetmaxproto.err_handler = jetmax_err;
	jetmaxproto.no_policy = 1;
	inet_add_protocol(&jetmaxproto,PROTOCOL);
	curr_queue_xmit = tf->queue_xmit;
	tf->queue_xmit = jetmax_to_ip_pkt_dispatcher;
	tf->net_header_len += sizeof(struct jetmaxhdr);
	netheader_len = tf->net_header_len; //Has a value of 52 bytes
	
	for (i = 0; i < 1024; i++) 
	{
		struct perflow* a = &(flows[i]);
		a->hash = i;
		a->to = NULL;
		a->bw_data = 0;
		a->is_bw_timer_up = 0;
		a->bw_last_update = 0;
	}
	printk( KERN_DEBUG "Module jetmax_host init\n" );
	return 0;
}

static void jetmax_host_exit_module(void)
{
	struct tcp_func* tf = &ipv4_specific;	
	inet_del_protocol(&jetmaxproto,PROTOCOL);	
	tf->queue_xmit = curr_queue_xmit;
	tf->net_header_len -= sizeof(struct jetmaxhdr); 
	printk( KERN_DEBUG "Module jetmax_host exit\n" );
}

module_init(jetmax_host_init_module);
module_exit(jetmax_host_exit_module);
