#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/stat.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/spinlock.h>
#include <net/protocol.h>
#include <net/pkt_sched.h>
#include <net/ip.h>

MODULE_DESCRIPTION("JetMax router module");
MODULE_AUTHOR("Saurabh Jain (saujain@tamu.edu)");
MODULE_LICENSE("GPL");

#define MAX_DEVICES 8

long double MAX(long double a1,long double a2)
{
	if(a1 > a2) return a1;
	else return a2;
}

long double MIN(long double a1,long double a2)
{
	if(a1 < a2) return a1;
	else return a1;
}

long double ABS(long double a1)
{
	if(a1 < 0) return (-1)*a1;
	else return a1;
}


struct jetmaxhdr {
	u8 A:1;         // Approve or disapprove
	u8 I:1;         // Interval
	u8 A_F:1;	// Approve or disapprove feedback
	u8 I_F:1;	// Interval feedback
	u8 U:4;         // Unused
	u8 RT;
	u8 RC;
	u8 RS;
	float pkt_loss;
	float fair_rate;
	float proposed_pktsize;
	float ipi;
	u8 RS_feedback;
	float pkt_loss_feedback;
	float fair_rate_feedback;
};

static unsigned long CAPACITY[] = {
				   121000,
				   121000,
				   121000,
				   121000,
				   121000,
				   121000,
				   121000,
				   121000
				  }; //B/msec

static int count = 8;
static int PROTOCOL = 202;
static int INTERVAL = 100; //msec
static int GAMMA = 100;
static int GAMMA_SCALE = 100;

module_param_array(CAPACITY,ulong,&count,S_IRUGO|S_IWUSR);
module_param(PROTOCOL,int,S_IRUGO);
module_param(INTERVAL,int,S_IRUGO);
module_param(GAMMA,int,S_IRUGO);
module_param(GAMMA_SCALE,int,S_IRUGO);


struct jetmax_router_data {
	double gamma;
	unsigned long y_l_;
	unsigned long w_l_;
	float g_l_;
	float p_l_;
	double sum_ipi_;	
	float n_l_;
	double min_diff_;	
	unsigned long input_traffic_;
	unsigned long capacity_;
	uint32_t pkt_loss_seq_;
	uint32_t control_timeout_;
	uint32_t min_queue_size_;
	uint32_t max_queue_size_;
	uint32_t curr_queue_size_;
	uint8_t is_control_timer_up_;
	uint8_t interval_bit_;
	uint8_t approval_;
	struct timer_list control_timer_;
	rwlock_t lk;
	uint8_t hash;
};

struct nf_hook_ops netfilter_router_out;
struct jetmax_router_data jetmax_router_priv[MAX_DEVICES];

unsigned int hash_name(char * name) {
	if(name[0]=='l') return 0;
	if(name[0]=='e' && name[3]=='0') return 1;
	if(name[0]=='e' && name[3]=='1') return 2;
	if(name[0]=='e' && name[3]=='2') return 3;
	if(name[0]=='e' && name[3]=='3') return 4;
	if(name[0]=='e' && name[3]=='4') return 5;
	if(name[0]=='e' && name[3]=='5') return 6;
	if(name[0]=='s' && name[3]=='0') return 7;
	return 0;
}

void control_timeout_handler(unsigned long data)
{
	int cpu;
	unsigned long input_bw = 0;
	unsigned long resp_bw = 0;
	unsigned long unresp_bw = 0;
	float n_l_ = 0;
	struct jetmax_router_data *str = (struct jetmax_router_data*) data;

	if(str==NULL) {
		printk(KERN_CRIT "#timeout handler: str cannot be null.\n");
		return;
	}

	cpu = get_cpu();
	write_lock(&(str->lk));

	del_timer_sync(&(str->control_timer_));
	str->is_control_timer_up_ = 0;

	input_bw = str->y_l_/(str->control_timeout_);
	resp_bw = str->w_l_/(str->control_timeout_);
	unresp_bw = (input_bw - resp_bw);
	n_l_ = (str->sum_ipi_)/(double)(str->control_timeout_);
	str->n_l_ = (int)(n_l_ + 0.5);
	
	if(input_bw != 0)
		str->p_l_ = (input_bw - str->gamma*str->capacity_)/(long double)input_bw;
	else
		str->p_l_ = 2;

	if(str->n_l_ > 0)
		 str->g_l_ = (str->gamma*str->capacity_ - unresp_bw)/str->n_l_;
	//else
	//	str->g_l_ = 0;
		
	
	if(input_bw > str->capacity_)
		str->approval_ = 0;
	else
		str->approval_ = 1;

	if(str->interval_bit_ == 1) 
		str->interval_bit_ = 0;
	else if(str->interval_bit_ == 0)
		str->interval_bit_ = 1;

	str->pkt_loss_seq_++;
 
	if(input_bw != 0 && str->hash == 3) {
		printk(KERN_INFO "%ld:%d:%ld,%ld,%ld,%ld,%ld,%d,%d,%d,%d,%d,%d,%d\n",
		jiffies,
		str->hash,(long)str->y_l_,(long)input_bw,(long)resp_bw,
		(long)unresp_bw,(long)str->input_traffic_,
		(int)(n_l_*1000),(int)str->n_l_,
	  (int)str->g_l_,(int)str->sum_ipi_,(int)str->approval_,str->curr_queue_size_,
		(int)(1000*str->p_l_));
	}


	str->y_l_ = 0;
	str->w_l_ = 0;
	str->sum_ipi_ = 0;
	str->n_l_ = 0;
	str->input_traffic_ = 0;

	init_timer(&(str->control_timer_));
	str->control_timer_.expires = (jiffies + (unsigned long)(((str->control_timeout_)*HZ)/1000.0));
	str->control_timer_.data = (unsigned long)str;
	str->control_timer_.function = control_timeout_handler;
	add_timer(&(str->control_timer_));
	str->is_control_timer_up_ = 1;

	write_unlock(&(str->lk));
	put_cpu();
}

unsigned int pkt_handler(unsigned int hooknum, struct sk_buff **skb,
			const struct net_device *in, const struct net_device *out,
			int (*okfn)(struct sk_buff*))
{
	unsigned int i = hash_name((char*)out->name);
	uint32_t packetsize = 0;
	uint32_t ip_header_len;
	int queuelen;
	double diff;
	struct jetmax_router_data *str;
	struct sk_buff *buff;
	struct iphdr *ihdr;
	struct jetmaxhdr *xhdr;

	if(i < 0 || i > MAX_DEVICES) 
		printk(KERN_CRIT "#pkt handler: invalid hash calculated: %d\n",i); 

	str = &jetmax_router_priv[i];
	str->hash = i;
	buff = *skb;
	ihdr = (struct iphdr*)buff->nh.iph;
	ip_header_len = (ihdr->ihl) * 4;

	if(ihdr->protocol == PROTOCOL) {
		xhdr = (struct jetmaxhdr*)((buff->data) + ip_header_len);
		packetsize = xhdr->proposed_pktsize;
		write_lock(&(str->lk));
		str->input_traffic_ += buff->len; //returns the IP level length (1500)
		str->y_l_ += packetsize;
		xhdr->RC = (xhdr->RC+1);

		if(xhdr->A != 0 && str->approval_ == 0) {
			xhdr->A = str->approval_;
		}

		if(xhdr->RC == xhdr->RT) {
			str->w_l_ += packetsize;
			str->sum_ipi_ += xhdr->ipi;
			xhdr->I = str->interval_bit_;
		}

		diff = ABS(str->p_l_  - xhdr->pkt_loss);

		if((xhdr->pkt_loss < str->p_l_ || xhdr->pkt_loss == 2) &&  
				str->p_l_ != 2 && diff > str->min_diff_) {
			if(str->g_l_ != 0) xhdr->fair_rate = str->g_l_;
			xhdr->RS = xhdr->RC;
			xhdr->pkt_loss = str->p_l_;
		} 
		else if((str->p_l_ == 2 && xhdr->pkt_loss == 2) || diff < str->min_diff_) {
			if(str->g_l_ !=0) {
				if(xhdr->fair_rate > 0) {
					if(str->g_l_ < xhdr->fair_rate 
						&& ABS(str->g_l_  - xhdr->fair_rate) > str->min_diff_) {
						xhdr->RS = xhdr->RC;
						xhdr->fair_rate = str->g_l_;
						xhdr->pkt_loss = str->p_l_;
					}
				} else {
					xhdr->RS = xhdr->RC;
					xhdr->fair_rate = str->g_l_;
					xhdr->pkt_loss = str->p_l_;
				}
			}
		}

		if((out != 0) && (out->qdisc != 0)) {
			queuelen = out->qdisc->q.qlen;
			if(queuelen < str->min_queue_size_) str->min_queue_size_ = queuelen;
			str->max_queue_size_ = MAX(str->max_queue_size_,queuelen);
			str->curr_queue_size_ = queuelen;
		}
		write_unlock(&(str->lk));
	}
	return NF_ACCEPT;
}

static int jetmaxr_init_module(void)
{
	int i = 0;
	netfilter_router_out.hook     = pkt_handler;
	netfilter_router_out.pf       = PF_INET;
	netfilter_router_out.hooknum  = NF_IP_POST_ROUTING;
	netfilter_router_out.priority = NF_IP_PRI_LAST;
	nf_register_hook(&netfilter_router_out);

	for (i = 0; i < MAX_DEVICES; i++) 
	{
		rwlock_init(&(jetmax_router_priv[i].lk));
		jetmax_router_priv[i].y_l_ = 0;
		jetmax_router_priv[i].w_l_ = 0;
		jetmax_router_priv[i].g_l_ = 0;
		jetmax_router_priv[i].p_l_ = 2;
		jetmax_router_priv[i].n_l_ = 0;
		jetmax_router_priv[i].pkt_loss_seq_ = 0;
		jetmax_router_priv[i].sum_ipi_ = 0;
		jetmax_router_priv[i].input_traffic_ = 0;
		jetmax_router_priv[i].control_timeout_ = INTERVAL;
		get_cpu();
		jetmax_router_priv[i].min_diff_ = 0.1;
		jetmax_router_priv[i].gamma = (double)GAMMA/(double)GAMMA_SCALE;
		put_cpu();
		jetmax_router_priv[i].min_queue_size_ = 0x0fffffff;	
		jetmax_router_priv[i].max_queue_size_ = 0;	
		init_timer(&(jetmax_router_priv[i].control_timer_));
		jetmax_router_priv[i].control_timer_.expires = (jiffies + (INTERVAL * HZ)/1000);
		jetmax_router_priv[i].control_timer_.data = (unsigned long)&jetmax_router_priv[i];
		jetmax_router_priv[i].control_timer_.function = control_timeout_handler;
		add_timer(&(jetmax_router_priv[i].control_timer_));
		jetmax_router_priv[i].is_control_timer_up_ = 1;
		jetmax_router_priv[i].interval_bit_ = 0;
		jetmax_router_priv[i].approval_ = 1;
		jetmax_router_priv[i].capacity_ = CAPACITY[i];
		
	}
	printk( KERN_INFO "Module jetmaxr init\n" );
	return 0;
}

static void jetmaxr_exit_module(void)
{
	int i = 0;
	nf_unregister_hook(&netfilter_router_out);
	for (i = 0; i < MAX_DEVICES; i++) {
		if (jetmax_router_priv[i].is_control_timer_up_==1) {
			jetmax_router_priv[i].is_control_timer_up_ = 0;
			del_timer_sync(&(jetmax_router_priv[i].control_timer_));
		}
	}
	printk( KERN_INFO "Module jetmaxr exit\n" );
}

module_init(jetmaxr_init_module);
module_exit(jetmaxr_exit_module);
