#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/skbuff.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/tcp.h>
#include <linux/types.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/jiffies.h>

MODULE_DESCRIPTION("My kernel module");
MODULE_AUTHOR("Saurabh Jain (saujain@tamu.edu)");
MODULE_LICENSE("GPL");

typedef union {
	double d;
	struct {
		#ifdef LITTLE_ENDIAN
			int j,i;
		#else
			int i,j;
		#endif
	} n;
} eco;

#define A 1512775.395
#define B 1072693248
#define C 60801

struct maxnethdr {
  uint8_t proto;
  float forward_price;	//For itself
  float reverse_price;	//To be feedback to the sender
} maxnethdr_t;

struct perflow {
  int hash;
  struct tcp_sock *to;
  float eta;	 //State Variable
  uint16_t cwnd; //Used at the time of loss to immediately ramp-up.
  float forward_price;
  float reverse_price;
};


static int PROTOCOL = 203;
static int P1 = 6;
static int P2 = 1;
static int P3 = 2;
static int P4 = 15;
static int INTERVAL = 10;
static int MAX_INCREMENT = 2;
static int MAX_DECREMENT = 20;
static struct net_protocol maxnetproto;
static struct perflow flows[1024];
int (*curr_queue_xmit) (struct sk_buff* skb, int ipfragok);

module_param(PROTOCOL,int,S_IRUGO);
module_param(P1,int,S_IRUGO);
module_param(P2,int,S_IRUGO);
module_param(P3,int,S_IRUGO);
module_param(P4,int,S_IRUGO);
module_param(INTERVAL,int,S_IRUGO);
module_param(MAX_INCREMENT,int,S_IRUGO);
module_param(MAX_DECREMENT,int,S_IRUGO);

extern struct net_protocol tcp_protocol;
extern struct tcp_func ipv4_specific;

struct perflow* get_flow(uint32_t daddr, uint32_t saddr, uint16_t dport, uint16_t sport) 
{
	unsigned char c = 0;
	unsigned int h = (saddr ^ sport) ^ (daddr ^ dport);
	h ^= h >> 16;
	h ^= h >> 8;
	c = (unsigned char)h;
	flows[c].hash = c;
	return &(flows[c]);
}

//IP->maxnet->TCP
//Possible Cases: 1) Receiver: receives a Data packet 
//                2) Sender: receives an Ack packet
//                3) Sender: receives a Data-Ack packet
int maxnet_pkt_handler(struct sk_buff* skb) 
{
	struct perflow* pd = 0;
	int cpu;
	float forward_price = 0, reverse_price = 0;
	double rtt, param;
	u16 tmpwnd, tmp;
	eco e1;
	struct maxnethdr* hdr = (struct maxnethdr*)skb->data;

	reverse_price = hdr->forward_price;		//This has to be fed back
	forward_price = hdr->reverse_price;		//This is for you

	if ((((hdr->proto) & (MAX_INET_PROTOS -1))) != 6) 
	{
		printk(KERN_CRIT "#maxnet_pkt_handler: Unknown protocol to call next.\n");
		return 0;
	}

	skb_pull(skb,sizeof(struct maxnethdr));
	skb->h.th = (struct tcphdr*)skb->data;

	pd = get_flow(skb->nh.iph->daddr,skb->nh.iph->saddr,
		ntohs((short)(skb->h.th->dest)),
		ntohs((short)(skb->h.th->source)));
		

	if(pd==NULL) {
		printk(KERN_CRIT "#maxnet_pkt_handler: Null value of pd obtained.\n");
		return 0;
	}

	cpu = get_cpu();

	pd->reverse_price = reverse_price;
	pd->forward_price = forward_price;
	
	if(pd->cwnd !=0 && (pd->to->rcv_rtt_est.rtt) !=0) {
		rtt = pd->to->rcv_rtt_est.rtt;
		pd->eta = pd->eta + (double)P1/(double)pd->cwnd - (double)(P2*pd->forward_price*INTERVAL)/(double)(rtt);
		param = (pd->eta - (double)(P4*pd->forward_price)/(double)rtt);
		e1.n.j = 0;
		e1.n.i = A*param + (B-C);
		
		tmpwnd = (u16) P3*e1.d;
		tmp = (pd->cwnd - tmpwnd); 
		
		if(tmp > MAX_DECREMENT)
			pd->cwnd = pd->cwnd - MAX_DECREMENT;
		else if(tmp < -MAX_INCREMENT)
			pd->cwnd = pd->cwnd + MAX_INCREMENT;
		else
			pd->cwnd = (u16)tmpwnd;
		
		if(pd->to!=NULL) {
			pd->to->snd_cwnd = pd->cwnd;
			pd->to->snd_ssthresh = (pd->to->snd_cwnd)/2;
			pd->to->snd_cwnd_stamp = tcp_time_stamp;
			pd->to->snd_cwnd_cnt = 0;
			printk(KERN_INFO "%d:%ld:%d:%ld:%d\n",pd->hash,jiffies,(int)(pd->to->srtt)/8,(long)(pd->to->snd_cwnd),
			   (int)(pd->to->ca_state));
		}
	put_cpu();
  }
  return tcp_protocol.handler(skb);
}

//TCP->maxnet->IP
//Possible cases: 1) Sender: Sending a data packet 
//		  2) Receiver: Sending an Ack 
//		  3) Receiver: Sending a Data-Ack packet
int maxnet_to_ip_pkt_dispatcher(struct sk_buff* skb, int ipfragok) 
{
	int cpu = 0;
	struct maxnethdr* hdr = 0;
	struct perflow* pd = 0;
	struct inet_sock* io = inet_sk(skb->sk);
	struct tcphdr* thdr = (struct tcphdr*)skb->h.th;

	pd = get_flow(io->daddr,io->saddr,ntohs((short)(thdr->dest)),
		ntohs((short)(thdr->source)));

	if(pd==NULL) {
		printk(KERN_CRIT "#Host-dispatcher: Null value of pd obtained.\n");
		return 0;
	}

	pd->to = tcp_sk(skb->sk);
	cpu = get_cpu();
	hdr = (struct maxnethdr*) skb_push(skb,sizeof(struct maxnethdr));
	hdr->proto = 6;
	hdr->forward_price = 0;
	hdr->reverse_price = pd->reverse_price;
	skb->sk->sk_protocol = PROTOCOL;
	put_cpu();
	return curr_queue_xmit(skb,ipfragok);
}

void maxnet_err (struct sk_buff* skb, u32 info) 
{
	printk("#MaxNet error handler called : %u\n",info);
	tcp_protocol.err_handler(skb,info);
}

static int maxnet_host_init_module(void)
{
	int i =0;
	struct tcp_func* tf = &ipv4_specific;
	maxnetproto.handler = maxnet_pkt_handler;
	maxnetproto.err_handler = maxnet_err;
	maxnetproto.no_policy = 1;
	inet_add_protocol(&maxnetproto,PROTOCOL);
	curr_queue_xmit = tf->queue_xmit;
	tf->queue_xmit = maxnet_to_ip_pkt_dispatcher;
	tf->net_header_len += sizeof(struct maxnethdr);	
	
	for (i = 0; i < 256; i++) 
	{
		struct perflow* a = &(flows[i]);
		a->hash = i;
		a->to = NULL;
		a->eta = 0;
		a->cwnd = 1;
		a->forward_price = 0;
		a->reverse_price = 0;
	}
	printk( KERN_ALERT "Module MaxNet_host init\n" );
	return 0;
}

static void maxnet_host_exit_module(void)
{
	struct tcp_func* tf = &ipv4_specific;	
	inet_del_protocol(&maxnetproto,PROTOCOL);	
	tf->queue_xmit = curr_queue_xmit;
	tf->net_header_len -= sizeof(struct maxnethdr); 
	printk( KERN_ALERT "Module MaxNet_host exit\n" );
}

module_init(maxnet_host_init_module);
module_exit(maxnet_host_exit_module);
