#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/stat.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/spinlock.h>
#include <net/protocol.h>
#include <net/pkt_sched.h>
#include <net/ip.h>

MODULE_DESCRIPTION("RCP Router Module");
MODULE_AUTHOR("Saurabh Jain (saujain@tamu.edu)");
MODULE_LICENSE("GPL");


#define MAX_DEVICES 8

long double MAX(long double a1,long double a2)
{
	if(a1 > a2) return a1;
	else return a2;
}

long double MIN(long double a1,long double a2)
{
	if(a1 < a2) return a1;
	else return a1;
}

long double ABS(long double a1)
{
	if(a1 < 0) return (-1)*a1;
	else return a1;
}

struct rcphdr {
	float fair_rate_;
	float fair_rate_feedback_;
	float rtt_;
};

static unsigned long CAPACITY[] = {
				   121000,
				   121000,
				   121000,
				   121000,
				   100000,
				   121000,
				   121000,
				   121000
				  }; //B/msec

static int count = 8;
static int PROTOCOL = 203;
static int MTU = 1500;
static int INTERVAL = 10; //msec
static double ALPHA = 0.1;
static double BETA = 1;
static double GAIN = 0.02;
static int alpha = 100;
static int beta = 1000;
static int gain = 2;

module_param_array(CAPACITY,ulong,&count,S_IRUGO|S_IWUSR);
module_param(PROTOCOL,int,S_IRUGO);
module_param(MTU,int,S_IRUGO);
module_param(INTERVAL,int,S_IRUGO);
module_param(alpha,int,S_IRUGO);
module_param(beta,int,S_IRUGO);
module_param(gain,int,S_IRUGO);


struct rcp_router_data {
	long double input_traffic_;
	unsigned long capacity_;
	uint32_t control_timeout_;
	uint32_t curr_queue_size_;
	uint32_t min_queue_size_;
	uint32_t max_queue_size_;
	float avg_rtt_;
	float fair_rate_;
	uint8_t is_control_timer_up_;
	struct timer_list control_timer_;
	rwlock_t lk;
	uint8_t hash;
};

struct nf_hook_ops netfilter_router_out;
struct rcp_router_data rcp_router_priv[MAX_DEVICES];

unsigned int hash_name(char * name) {
	if(name[0]=='l') return 0;
	if(name[0]=='e' && name[3]=='0') return 1;
	if(name[0]=='e' && name[3]=='1') return 2;
	if(name[0]=='e' && name[3]=='2') return 3;
	if(name[0]=='e' && name[3]=='3') return 4;
	if(name[0]=='e' && name[3]=='4') return 5;
	if(name[0]=='e' && name[3]=='5') return 6;
	if(name[0]=='s' && name[3]=='0') return 7;
	return 0;
}

void control_timeout_handler(unsigned long data)
{
	int cpu;
	long double input_bw = 0;
	struct rcp_router_data *str = (struct rcp_router_data*) data;
	double numerator = 0, denominator = 0;
	double numflows = 0;

	if(str == NULL) {
		printk(KERN_CRIT "#timeout handler: str cannot be null.\n");
		return;
	}

	cpu = get_cpu();
	write_lock(&(str->lk));
	del_timer_sync(&(str->control_timer_));
	str->is_control_timer_up_ = 0;

	if(str->avg_rtt_ != 0 && str->input_traffic_ != 0) 
	{
		input_bw = (double)(str->input_traffic_)/(double)(str->control_timeout_);
		numerator = str->control_timeout_*(ALPHA*(str->capacity_ - input_bw) - 
								BETA*((double)(MTU*str->curr_queue_size_)/(double)str->avg_rtt_));
		denominator = str->capacity_ * str->avg_rtt_;

		if(str->fair_rate_) {
			numflows = str->capacity_/str->fair_rate_;
		}

		str->fair_rate_ = str->fair_rate_ * (1 + (double)numerator/(double)denominator);

		if(str->fair_rate_ > str->capacity_) str->fair_rate_ = str->capacity_;

		if(str->hash == 4) {
			printk(KERN_INFO "%ld:%ld,%ld,%ld,%ld,%ld,%d,%d\n", 
									jiffies,(long int)(str->input_traffic_),
									(long int)str->curr_queue_size_, (long int)str->fair_rate_,
									(unsigned long)str->avg_rtt_,(long)(1000*numflows),
									str->min_queue_size_, str->max_queue_size_);
		}
	}
	str->input_traffic_ = 0;
	str->max_queue_size_ = 0;
	str->min_queue_size_ = 0x0fffffff;
	init_timer(&(str->control_timer_));
	str->control_timer_.expires = (jiffies + 
											(unsigned long)(((str->control_timeout_)*HZ)/1000.0));
	str->control_timer_.data = (unsigned long)str;
	str->control_timer_.function = control_timeout_handler;
	add_timer(&(str->control_timer_));
	str->is_control_timer_up_ = 1;
	write_unlock(&(str->lk));
	put_cpu();
}

unsigned int pkt_handler(unsigned int hooknum, struct sk_buff **skb,
			const struct net_device *in, const struct net_device *out,
			int (*okfn)(struct sk_buff*))
{
	unsigned int i=hash_name((char*)out->name);
	uint32_t packetsize = 0;
	uint32_t queuelen = 0; 
	uint32_t ip_header_len;
	int cpu;
	struct rcp_router_data *str;
	struct sk_buff *buff;
	struct iphdr *ihdr;
	struct rcphdr *xhdr;

	if(i < 0 || i > MAX_DEVICES) 
		printk(KERN_CRIT "#pkt handler: invalid hash calculated: %d\n",i); 

	str = &rcp_router_priv[i];
	str->hash = i;
	buff = *skb;
	ihdr = (struct iphdr*)buff->nh.iph;
	ip_header_len = (ihdr->ihl) * 4;

	if(ihdr->protocol==PROTOCOL) {

		xhdr = (struct rcphdr*)((buff->data) + ip_header_len);
		packetsize = (buff->len);
		cpu = get_cpu();
		write_lock(&(str->lk));
		str->input_traffic_ +=packetsize;

		if(xhdr->rtt_ > 0 && xhdr->rtt_ < 5000) {
			str->avg_rtt_ = GAIN*xhdr->rtt_ + (1 - GAIN)*str->avg_rtt_;
		}

		if(xhdr->fair_rate_ > str->fair_rate_ || xhdr->fair_rate_ == -1) 
			xhdr->fair_rate_ = str->fair_rate_;


		if((out != 0) && (out->qdisc != 0)) {
			queuelen = out->qdisc->q.qlen;
			if(queuelen < str->min_queue_size_) str->min_queue_size_ = queuelen;
			str->max_queue_size_ = MAX(str->max_queue_size_,queuelen);
			str->curr_queue_size_ = queuelen;
		}

		write_unlock(&(str->lk));

		put_cpu();
		}
	return NF_ACCEPT;
}


static int rcpr_init_module(void)
{
	int i = 0;
	ALPHA = ((double)alpha)/1000.0;
	BETA = ((double)beta)/1000.0;
	GAIN = ((double)gain)/100.0;
	netfilter_router_out.hook     = pkt_handler;
	netfilter_router_out.pf       = PF_INET;
	netfilter_router_out.hooknum  = NF_IP_POST_ROUTING;
	netfilter_router_out.priority = NF_IP_PRI_LAST;
	nf_register_hook(&netfilter_router_out);

	for (i = 0; i < MAX_DEVICES; i++) 
	{
		rwlock_init(&(rcp_router_priv[i].lk));
		rcp_router_priv[i].input_traffic_ = 0;
		rcp_router_priv[i].capacity_ = CAPACITY[i];
		rcp_router_priv[i].control_timeout_ = INTERVAL;
		rcp_router_priv[i].min_queue_size_ = 0x0fffffff;
		rcp_router_priv[i].max_queue_size_ = 0;
		rcp_router_priv[i].curr_queue_size_ = 0;
		rcp_router_priv[i].avg_rtt_ = 0;
		rcp_router_priv[i].fair_rate_ = 125;
		init_timer(&(rcp_router_priv[i].control_timer_));
		rcp_router_priv[i].control_timer_.expires = (jiffies + 
																								(INTERVAL * HZ)/1000);
		rcp_router_priv[i].control_timer_.data = (unsigned long)&rcp_router_priv[i];
		rcp_router_priv[i].control_timer_.function = control_timeout_handler;
		add_timer(&(rcp_router_priv[i].control_timer_));
		rcp_router_priv[i].is_control_timer_up_ = 1;
	}
	printk( KERN_INFO "Module rcpr init\n" );
	return 0;
}

static void rcpr_exit_module(void)
{
	int i = 0;
	nf_unregister_hook(&netfilter_router_out);
	for (i = 0; i < MAX_DEVICES; i++) {
		if (rcp_router_priv[i].is_control_timer_up_==1) {
			rcp_router_priv[i].is_control_timer_up_ = 0;
			del_timer_sync(&(rcp_router_priv[i].control_timer_));
		}
	}
	printk( KERN_INFO "Module rcpr exit\n" );
}

module_init(rcpr_init_module);
module_exit(rcpr_exit_module);
