#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/skbuff.h>
#include <linux/types.h>
#include <linux/stat.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <net/tcp.h>
#include <linux/types.h>

#define MAX_FLOWS 256

MODULE_DESCRIPTION("XCP Host");
MODULE_AUTHOR("saujain (saujain@tamu.edu)");
MODULE_LICENSE("$LICENSE$");

#define MAX(x,y) (x > y ? x : y)
#define MIN(x,y) (x < y ? x : y)
#define ABS(x) (x > 0 ? x : (-1)*(x))

struct xcphdr {
  uint8_t proto;
  uint8_t len;
  uint8_t verform;
  uint8_t unused;
  double rtt;  
  double x;
  double delta;
  double reverse;
};

struct perflow {
  uint8_t hash;
  double rtt;  
  double x;
  double delta;
  double reverse;
};

static unsigned long CAPACITY = 125000000; //B/s
static int PROTOCOL = 6000;
static int MSS =1500;
static unsigned long DESIRED_RATE=125000000; //B/s
static struct net_protocol xcpproto;
static struct perflow flows[256];
int (*curr_queue_xmit) (struct sk_buff* skb, int ipfragok);

module_param(CAPACITY,ulong,S_IRUGO);
module_param(PROTOCOL,int,S_IRUGO);
module_param(MSS,int,S_IRUGO);

extern struct net_protocol tcp_protocol;  
extern struct tcp_func ipv4_specific;

struct perflow* get_flow(uint32_t daddr, uint32_t saddr, uint16_t dport, uint16_t sport) 
{
  uint16_t h = (saddr ^ sport) ^ (daddr ^ dport);
  h ^= h >> 16;
  h ^= h >> 8;
  flows[h].hash = h;
  return &(flows[h]);
}

//IP->XCP->TCP
//Possible Cases: 1) Receiver: receives a data packet, 2) Sender: receives an ack packet
int xcp_pkt_handler(struct sk_buff* skb) 
{
  double rtt;  
  double x;
  double delta;
  double reverse;
  double increment;	
  int saddr   = 0;
  int daddr   = 0;
  int sport   = 0;
  int dport   = 0;
  
  struct perflow* pd = 0;
  struct xcphdr* hdr = (struct xcphdr*)skb->data;
  struct tcphdr* thdr;
  struct tcp_opt* to = tcp_sk(skb->sk);
  
  saddr = skb->nh.iph->saddr;
  daddr = skb->nh.iph->daddr;
  sport = skb->h.th->source;
  dport = skb->h.th->dest;
  
  rtt = hdr->rtt;
  x = hdr->x;
  delta = hdr->delta;
  reverse = hdr->reverse; 
  
  pd = get_flow(saddr,sport,daddr,dport);
  
  skb_pull(skb,hdr->len);
  skb->h.th = (struct tcphdr*)skb->data;
  thdr = (struct tcphdr*) skb->h.th;
  
  if(thdr->ack==1) {
  //This is an ack. Update the tcp stack.
	increment = (reverse*rtt)/MSS;
	to->snd_cwnd = MAX((uint32_t)(to->snd_cwnd + increment) , 1);
	to->snd_cwnd_used = MAX((uint32_t)(to->snd_cwnd_used + increment) , 1);
  } else {
  //We received a packet. Save the state to be sent in the ack.
  	pd->delta = delta;
  	pd->reverse = reverse;
  	pd->rtt = rtt;  
  	pd->x = x;
  }
  return tcp_protocol.handler(skb);
}

//TCP->XCP->IP
//Possible cases: 1) Sender: Sending a packet 2) Receiver: Sending an ack
int xcp_to_ip_pkt_dispatcher(struct sk_buff* skb, int ipfragok) 
{
  struct xcphdr* hdr = 0;
  struct perflow* pd = 0;
  struct tcp_opt* to = tcp_sk(skb->sk);
  struct inet_opt* io = inet_sk(skb->sk);
  struct tcphdr* thdr = (struct tcphdr*)skb->h.th;
  
  if (skb->sk->sk_state == 1) { // TCP - Established
    int sport = thdr->source;
    int dport = thdr->dest;
    int saddr = io->saddr;
    int daddr = io->daddr;
    
    pd = get_flow(saddr,sport,daddr,dport);
    hdr = (struct xcphdr*) skb_push(skb,20);
    hdr->proto = 6;
    hdr->len = sizeof(struct xcphdr);
    hdr->unused = 0;
    
    if(thdr->ack==1) {
    	//This is an ack. Update xcp header from flows data
    	hdr->verform = 0x12;
    	hdr->rtt = pd->rtt;
    	hdr->x = pd->x;
    	hdr->delta = pd->delta;
    	hdr->reverse = pd->reverse;
    } else {
    	//This is a new sender packet. Update xcp header with data from tcp stack
    	hdr->verform = 0x11;
	hdr->rtt = (to->srtt)/8;
	hdr->reverse = 0;
	if(hdr->rtt !=0) {
		hdr->x = (hdr->rtt)/((to->snd_cwnd)*MSS);
		hdr->delta = (DESIRED_RATE - ((to->snd_cwnd)*MSS)/(hdr->rtt));
	} else {
		hdr->x = 0;
		hdr->delta = DESIRED_RATE;
	}
    }
    skb->sk->sk_protocol = PROTOCOL;
    } else {
    	//TCP connection is not established.
        if (skb->sk->sk_protocol == PROTOCOL) {
    	  skb->sk->sk_protocol = 6;
    	}
  }
  return curr_queue_xmit(skb,ipfragok);
}    

void xcp_err (struct sk_buff* skb, u32 info) 
{
  printk("XPC error handler called : %u\n",info);
  tcp_protocol.err_handler(skb,info);
}

static int xcp_host_init_module(void)
{
	int i =0;
    	struct tcp_func* tf = &ipv4_specific;
	xcpproto.handler = xcp_pkt_handler;
  	xcpproto.err_handler = xcp_err;
  	xcpproto.no_policy = 1;
  	inet_add_protocol(&xcpproto,PROTOCOL);
    	curr_queue_xmit = tf->queue_xmit;
  	tf->queue_xmit = xcp_to_ip_pkt_dispatcher;
    	tf->net_header_len += sizeof(struct xcphdr);	
	
  	for (i = 0; i < 256; i++) 
	{
    		struct perflow* a = &(flows[i]);
    		a->hash = i;
    		a->rtt = 0;
		a->x = 0;
		a->delta = 0;
		a->reverse = 0;
  	}
    	
  	printk( KERN_ALERT "Module xcp_host init\n" );
	return 0;
}

static void xcp_host_exit_module(void)
{
	struct tcp_func* tf = &ipv4_specific;	
  	inet_del_protocol(&xcpproto,PROTOCOL);	
  	tf->queue_xmit = curr_queue_xmit;
  	tf->net_header_len -= sizeof(struct xcphdr); 
  	printk( KERN_ALERT "Module xcp_host exit done! \n" );
}

module_init(xcp_host_init_module);
module_exit(xcp_host_exit_module);
