#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/stat.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/spinlock.h>
#include <net/protocol.h>
#include <net/pkt_sched.h>
#include <net/ip.h>

MODULE_DESCRIPTION("XCP Router Module");
MODULE_AUTHOR("saujain (saujain@tamu.edu)");
MODULE_LICENSE("$LICENSE$");

#define MAX(x,y) (x > y ? x : y)
#define MIN(x,y) (x < y ? x : y)
#define ABS(x) (x > 0 ? x : (-1)*(x))
#define MAX_DEVICES 4

struct xcphdr {
  uint8_t proto;
  uint8_t len;
  uint8_t verform;
  uint8_t unused;
  double rtt;  
  double x;
  double delta;
  double reverse;
} xcphrd_t;

static unsigned long CAPACITY[] = { 125000000,125000000,125000000,125000000 } ; //B/s
static int count = 4;
static int PROTOCOL = 6000;
static int MSS =1500;
static double SHUFFLE_FACTOR=0.1;
static double ALPHA = 0.4;
static double BETA = 0.226;
static int shuffle_factor = 100;
static int alpha = 400;
static int beta = 226;
static int INITIAL_TIMEOUT = 1; //msec

module_param_array(CAPACITY,ulong,count,S_IRUGO|S_IWUSR);
module_param(PROTOCOL,int,S_IRUGO);
module_param(MSS,int,S_IRUGO);
module_param(alpha,int,S_IRUGO);
module_param(beta,int,S_IRUGO);
module_param(shuffle_factor,int,S_IRUGO);
module_param(INITIAL_TIMEOUT,int,S_IRUGO);

struct xcp_router_data {
	double input_traffic_;
	double avg_rtt_;
	double sum_rtt_by_cwnd_;
	double sum_rttsq_by_cwnd_;
	double epsilonp_;
	double epsilonn_;
	double residue_pos_fbk_;
	double residue_neg_fbk_;
	unsigned long capacity_;
	uint32_t min_queue_size_;
	uint32_t max_queue_size_;
	uint8_t is_timer_up_;
	struct timer_list control_timer_;
	rwlock_t lk;
} xcp_router_data_t;

struct nf_hook_ops netfilter_router_out;
struct xcp_router_data xcp_router_priv[MAX_DEVICES];

unsigned int hash_name(char * name) {
	unsigned int sum = 0;
	int i = 0;
	for(i=0; i < IFNAMSIZ; i++) {
		sum+=(unsigned int)name[i];
	}
	return (sum % MAX_DEVICES);
}

void timeout_handler(unsigned long data)
{
	double avg_rtt_ = 0;
	double phi = 0;
	double shuffled_traffic = 0;
	struct xcp_router_data *str = (struct xcp_router_data*) data;
	del_timer_sync(&(str->control_timer_));
	avg_rtt_ = ((str->sum_rttsq_by_cwnd_)/(str->sum_rtt_by_cwnd_)); 
	phi = ALPHA * (str->avg_rtt_) * ((str->capacity_) - (str->input_traffic_) / (str->avg_rtt_)) - 
		     BETA * (str->min_queue_size_) * MSS;
	shuffled_traffic = SHUFFLE_FACTOR * (str->input_traffic_) / avg_rtt_;
	write_lock(&(str->lk));
	str->epsilonp_ = ((max(phi,0.0) + shuffled_traffic)/(avg_rtt_ * (str->sum_rtt_by_cwnd_)));
	str->epsilonn_ = ((max(-phi,0.0) + shuffled_traffic)/(avg_rtt_ * (str->input_traffic_)));
	str->residue_pos_fbk_ = (max(phi,0.0) + shuffled_traffic)/avg_rtt_;
	str->residue_neg_fbk_ = (max(-phi,0.0) + shuffled_traffic)/avg_rtt_;
	str->avg_rtt_ = avg_rtt_;
	str->input_traffic_ = 0;
	str->sum_rtt_by_cwnd_ = 0;
	str->sum_rttsq_by_cwnd_ = 0;
	init_timer(&(str->control_timer_));
	str->control_timer_.expires = (jiffies + (unsigned long)(avg_rtt_ * HZ));
	str->control_timer_.data = (unsigned long)str;
	str->control_timer_.function = timeout_handler;
	add_timer(&(str->control_timer_));
	write_unlock(&(str->lk));
}

unsigned int pkt_handler(unsigned int hooknum, struct sk_buff **skb,
			const struct net_device *in, const struct net_device *out,
			int (*okfn)(struct sk_buff*))
{
	struct xcp_router_data *str = &xcp_router_priv[hash_name((char*)out->name)];
	struct sk_buff* buff = *skb;
	struct iphdr* ihdr = (struct iphdr*)buff->nh.iph;
	uint32_t ip_header_len = (ihdr->ihl) * 4;
	uint32_t packetsize = 0;
	uint32_t queuelen = 0; 
	double pos_fbk = 0;
	double neg_fbk = 0;
	double feedback = 0;
		
	if(ihdr->protocol==PROTOCOL) {
		
		struct xcphdr* xhdr = (struct xcphdr*)((buff->data) + ip_header_len);
		
		if(xhdr->verform ==0x11) {
			
			packetsize = ((buff->len) - (buff->mac_len));//This has a max value of 1500
			write_lock(&(str->lk));
			if(xhdr->rtt!=0 || xhdr->x!=0) {
				str->input_traffic_ +=packetsize;
				str->sum_rtt_by_cwnd_ += ((xhdr->x) * packetsize);
				str->sum_rttsq_by_cwnd_ += ((xhdr->x) * (xhdr->rtt) * packetsize);
			}
		
			pos_fbk = (str->epsilonp_) * (xhdr->rtt) * (xhdr->x) *packetsize;
			neg_fbk = (str->epsilonn_) * (xhdr->rtt) * packetsize;
			feedback = (pos_fbk - neg_fbk);	
			
			if(xhdr->reverse >= feedback) {
				xhdr->reverse = feedback;
				str->residue_pos_fbk_ -= pos_fbk/(xhdr->rtt);
				str->residue_neg_fbk_ -= neg_fbk/(xhdr->rtt);
			} else {
				if(xhdr->reverse >=0) {
					str->residue_pos_fbk_ -= (xhdr->reverse)/(xhdr->rtt);
					str->residue_neg_fbk_ -= (feedback-(xhdr->reverse))/(xhdr->rtt);
				} else {
					str->residue_neg_fbk_ +=(xhdr->reverse)/(xhdr->rtt);
					if(feedback >=0) {
						str->residue_neg_fbk_ -=feedback/(xhdr->rtt);
					}
				}
			}
			if(str->residue_pos_fbk_ <=0) str->epsilonp_ = 0;
			if(str->residue_neg_fbk_ <=0) str->epsilonn_ = 0;
			
			if((out != 0) && (out->qdisc != 0)) {
				queuelen = out->qdisc->q.qlen;
				if(queuelen < str->min_queue_size_) str->min_queue_size_ = queuelen;
				if(queuelen > str->max_queue_size_) str->max_queue_size_ = queuelen;
			}
			write_unlock(&(str->lk));
		}
	}
	return NF_ACCEPT;
}

static int xcpr_init_module(void)
{
	int i = 0;
	ALPHA = alpha/1000;
	BETA = beta/1000;
	SHUFFLE_FACTOR = shuffle_factor/1000;
	netfilter_router_out.hook     = pkt_handler;
	netfilter_router_out.pf       = PF_INET;
	netfilter_router_out.hooknum  = NF_IP_POST_ROUTING;
	netfilter_router_out.priority = NF_IP_PRI_LAST;
	nf_register_hook(&netfilter_router_out);

	for (i = 0; i < MAX_DEVICES; i++) 
	{
		rwlock_init(&(xcp_router_priv[i].lk));
		xcp_router_priv[i].input_traffic_ = 0;
		xcp_router_priv[i].avg_rtt_ = 0;
		xcp_router_priv[i].sum_rtt_by_cwnd_ = 0;
		xcp_router_priv[i].sum_rttsq_by_cwnd_ = 0;
		xcp_router_priv[i].epsilonp_ = 0;
		xcp_router_priv[i].epsilonn_ = 0;
		xcp_router_priv[i].residue_pos_fbk_ = 0;
		xcp_router_priv[i].residue_neg_fbk_ = 0;
		xcp_router_priv[i].min_queue_size_ = 0;	
		xcp_router_priv[i].max_queue_size_ = 0;
		init_timer(&(xcp_router_priv[i].control_timer_));
		xcp_router_priv[i].control_timer_.expires = (jiffies + INITIAL_TIMEOUT * HZ / 1000);
		xcp_router_priv[i].control_timer_.data = (unsigned long)&xcp_router_priv[i];
		xcp_router_priv[i].control_timer_.function = timeout_handler;
		add_timer(&(xcp_router_priv[i].control_timer_));
		xcp_router_priv[i].is_timer_up_ = 1;
		xcp_router_priv[i].capacity_ = CAPACITY[i];
		
	}
	printk( KERN_ALERT "Module xcpr init\n" );
	return 0;
}

static void xcpr_exit_module(void)
{
	int i = 0;
	nf_unregister_hook(&netfilter_router_out);
	for (i = 0; i < MAX_DEVICES; i++) {
		if (xcp_router_priv[i].is_timer_up_==1) {
			xcp_router_priv[i].is_timer_up_ = 0;
			del_timer_sync(&(xcp_router_priv[i].control_timer_));
		}
	}
	printk( KERN_ALERT "Module xcpr exit\n" );
}

module_init(xcpr_init_module);
module_exit(xcpr_exit_module);
