#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/stat.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/netdevice.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/spinlock.h>
#include <net/protocol.h>
#include <net/pkt_sched.h>
#include <net/ip.h>

MODULE_DESCRIPTION("XCP Router Module");
MODULE_AUTHOR("saujain (saujain@tamu.edu)");
MODULE_LICENSE("GPL");

#define MAX_DEVICES 8

long double MAX(long double a1,long double a2)
{
	if(a1 > a2) return a1;
	else return a2;
}

long double MIN(long double a1,long double a2)
{
	if(a1 < a2) return a1;
	else return a1;
}

long double ABS(long double a1)
{
	if(a1 < 0) return (-1)*a1;
	else return a1;
}


struct xcphdr {
  uint8_t proto;
  uint8_t len;
  uint8_t verform;
  uint8_t unused;
  float rtt;  
  float x;
  float delta;
  float reverse;
} xcphrd_t;

static unsigned long CAPACITY[] = {
				   121000,
				   121000,
				   121000,
				   121000,
				   121000,
				   121000,
				   121000,
				   121000
				  }; //B/msec

static int count = 8;
static int PROTOCOL = 201;
static int MTU =1500;
static double SHUFFLE_FACTOR=0.1;
static double ALPHA = 0.4;
static double BETA = 0.226;
static int shuffle_factor = 100;
static int alpha = 400;
static int beta = 226;
static int MIN_INTERVAL = 10; //msec
static int MAX_INTERVAL = 500; //msec
static int ALLOWED_QUEUE = 2; //msec
static int IS_R = 0;

module_param_array(CAPACITY,ulong,&count,S_IRUGO|S_IWUSR);
module_param(PROTOCOL,int,S_IRUGO);
module_param(MTU,int,S_IRUGO);
module_param(alpha,int,S_IRUGO);
module_param(beta,int,S_IRUGO);
module_param(shuffle_factor,int,S_IRUGO);
module_param(MIN_INTERVAL,int,S_IRUGO);
module_param(MAX_INTERVAL,int,S_IRUGO);
module_param(ALLOWED_QUEUE,int,S_IRUGO);
module_param(IS_R,int,S_IRUGO);

struct xcp_router_data {
	int hash;
	long double input_traffic_;
	double avg_rtt_;
	uint32_t control_timeout_;
	uint32_t queue_timeout_;
	long double sum_rtt_by_cwnd_;
	long double sum_rttsq_by_cwnd_;
	double epsilonp_;
	double epsilonn_;
	long double residue_pos_fbk_;
	long double residue_neg_fbk_;
	unsigned long capacity_;
	uint32_t min_queue_size_;
	uint32_t max_queue_size_;
	uint32_t min_persistent_queue_;
	uint32_t curr_queue_size_;
	uint8_t is_control_timer_up_;
	uint8_t is_queue_timer_up_;
	struct timer_list control_timer_;
	struct timer_list queue_timer_;
	rwlock_t lk;
} xcp_router_data_t;

struct nf_hook_ops netfilter_router_out;
struct xcp_router_data xcp_router_priv[MAX_DEVICES];

unsigned int hash_name(char * name) {
	if(name[0]=='l') return 0;
	if(name[0]=='e' && name[3]=='0') return 1;
	if(name[0]=='e' && name[3]=='1') return 2;
	if(name[0]=='e' && name[3]=='2') return 3;
	if(name[0]=='e' && name[3]=='3') return 4;
	if(name[0]=='e' && name[3]=='4') return 5;
	if(name[0]=='e' && name[3]=='5') return 6;
	if(name[0]=='s' && name[3]=='0') return 7;
	return 0;
}

void control_timeout_handler(unsigned long data)
{
	int cpu;
	double avg_rtt_ = 0;
	long double phi = 0;
	long double shuffled_traffic = 0;
	long double input_bw = 0;
	struct xcp_router_data *str = (struct xcp_router_data*) data;

	if(str==NULL) {
		printk(KERN_CRIT "#timeout handler: str cannot be null.\n");
		return;
	}

	cpu = get_cpu();
	write_lock(&(str->lk));
	del_timer_sync(&(str->control_timer_));
	str->is_control_timer_up_ = 0;

	if(str->sum_rtt_by_cwnd_ !=0 && str->sum_rttsq_by_cwnd_ !=0) 
	{
		avg_rtt_ = ((str->sum_rttsq_by_cwnd_)/(str->sum_rtt_by_cwnd_));
		input_bw = (str->input_traffic_)/(str->control_timeout_);

		phi = ALPHA * (str->capacity_ - input_bw) - BETA * ((long double)str->min_persistent_queue_*MTU)/avg_rtt_;
		

		shuffled_traffic = MAX(0,(long double)(SHUFFLE_FACTOR * input_bw - ABS(phi)));
		str->residue_pos_fbk_ = (MAX(phi,0.0) + shuffled_traffic);
		str->residue_neg_fbk_ = (MAX(-phi,0.0) + shuffled_traffic);
		str->epsilonp_ = (str->residue_pos_fbk_)/(str->sum_rtt_by_cwnd_);
		str->epsilonn_ = (str->residue_neg_fbk_)/(str->input_traffic_);
		str->avg_rtt_ = avg_rtt_;

		/*
		if(str->hash == 6) {
			printk(KERN_INFO "%ld:%u:%ld:%ld:%ld:%ld:%ld\n",jiffies,
			(unsigned int)(str->capacity_),(long int)(str->input_traffic_),
			(long int)(str->control_timeout_), (long int)(str->min_persistent_queue_),
			(long int)(str->max_queue_size_), (long int)(str->curr_queue_size_));
		}
		*/
	}

	str->input_traffic_ = 0;
	str->sum_rtt_by_cwnd_ = 0;
	str->sum_rttsq_by_cwnd_ = 0;
	init_timer(&(str->control_timer_));

	if(avg_rtt_ < MIN_INTERVAL) str->control_timeout_ = MIN_INTERVAL;
	else str->control_timeout_ = avg_rtt_;

	str->control_timer_.expires = (jiffies + 
													(unsigned long)(((str->control_timeout_)*HZ)/1000.0));
	str->control_timer_.data = (unsigned long)str;
	str->control_timer_.function = control_timeout_handler;
	add_timer(&(str->control_timer_));
	str->is_control_timer_up_ = 1;

	write_unlock(&(str->lk));
	put_cpu();
}

void queue_timeout_handler(unsigned long data)
{
	int cpu = 0;
	struct xcp_router_data *str = (struct xcp_router_data*) data;
	cpu =  get_cpu();
	write_lock(&(str->lk));
	if(str->min_queue_size_==0x0fffffff) {
		str->min_queue_size_ = 0;
	}

	//printk(KERN_INFO "Min_queue_len=%d,Max_queue_len=%d\n",
	//					str->min_queue_size_,str->max_queue_size_);
	
	str->min_persistent_queue_=str->min_queue_size_;
	del_timer_sync(&(str->queue_timer_));
	str->is_queue_timer_up_ = 0;
	str->queue_timeout_ = MAX(ALLOWED_QUEUE,(str->avg_rtt_
												-(str->min_queue_size_/str->capacity_))/2); 
	init_timer(&(str->queue_timer_));

	str->queue_timer_.expires = (jiffies + 
														(unsigned long)(((str->queue_timeout_)*HZ)/1000.0));
	str->queue_timer_.data = (unsigned long)str;
	str->queue_timer_.function = queue_timeout_handler;
	str->min_queue_size_ = 0x0fffffff;
	str->max_queue_size_ = 0;
	add_timer(&(str->queue_timer_));
	str->is_queue_timer_up_ = 1;
	write_unlock(&(str->lk));
	put_cpu();
}

unsigned int pkt_handler(unsigned int hooknum, struct sk_buff **skb,
			const struct net_device *in, const struct net_device *out,
			int (*okfn)(struct sk_buff*))
{
	unsigned int i=hash_name((char*)out->name);
	uint32_t packetsize = 0;
	uint32_t queuelen = 0; 
	uint32_t ip_header_len;
	int cpu;
	double pos_fbk = 0;
	double neg_fbk = 0;
	double feedback = 0;
	struct xcp_router_data *str;
	struct sk_buff *buff;
	struct iphdr *ihdr;
	struct xcphdr *xhdr;
	if(i < 0 || i > MAX_DEVICES) 
		printk(KERN_CRIT "#pkt handler: invalid hash calculated: %d\n",i); 
	str = &xcp_router_priv[i];
	str->hash = i;
	buff = *skb;
	ihdr = (struct iphdr*)buff->nh.iph;
	ip_header_len = (ihdr->ihl) * 4;
	if(ihdr->protocol==PROTOCOL) {
		xhdr = (struct xcphdr*)((buff->data) + ip_header_len);
		if(xhdr->verform ==0x11) {
			packetsize = (buff->len);
			write_lock(&(str->lk));
			if(xhdr->rtt!=0 && xhdr->x!=0) 
			{
				cpu = get_cpu();
				str->input_traffic_ +=packetsize;
				str->sum_rtt_by_cwnd_ += (xhdr->x);
				if(xhdr->rtt < MAX_INTERVAL) {
					str->sum_rttsq_by_cwnd_ += ((xhdr->x) * (xhdr->rtt));
				} else {
					str->sum_rttsq_by_cwnd_ += ((xhdr->x) * MAX_INTERVAL);
				}	
				pos_fbk = (str->epsilonp_) * (xhdr->x);
				neg_fbk = (str->epsilonn_) * packetsize;
				feedback = (pos_fbk - neg_fbk);	
			
				if(xhdr->delta > feedback) {
					xhdr->delta = feedback;
				} else {
					neg_fbk = MIN(str->residue_neg_fbk_,neg_fbk+(feedback-(xhdr->delta)));
					pos_fbk = (xhdr->delta + neg_fbk);
				}
				str->residue_pos_fbk_ = MAX(0, str->residue_pos_fbk_-pos_fbk);
				str->residue_neg_fbk_ = MAX(0, str->residue_neg_fbk_-neg_fbk);

				if(str->residue_pos_fbk_ <=0) str->epsilonp_ = 0;
				if(str->residue_neg_fbk_ <=0) str->epsilonn_ = 0;
				put_cpu();
			}
			if((out != 0) && (out->qdisc != 0)) {
				queuelen = out->qdisc->q.qlen;
				if(queuelen < str->min_queue_size_) {
					str->min_queue_size_ = queuelen;
				}
				str->max_queue_size_ = MAX(str->max_queue_size_,queuelen);
				str->curr_queue_size_ = queuelen;
			}
			write_unlock(&(str->lk));
		}
	}
	return NF_ACCEPT;
}

static int xcpr_init_module(void)
{
	int i = 0;
	int cpu;
	cpu = get_cpu();
	ALPHA = ((double)alpha)/1000.0;
	BETA = ((double)beta)/1000.0;
	SHUFFLE_FACTOR = ((double)shuffle_factor)/1000.0;
	put_cpu();
	netfilter_router_out.hook     = pkt_handler;
	netfilter_router_out.pf       = PF_INET;
	netfilter_router_out.hooknum  = NF_IP_POST_ROUTING;
	netfilter_router_out.priority = NF_IP_PRI_LAST;
	nf_register_hook(&netfilter_router_out);

	for (i = 0; i < MAX_DEVICES; i++) 
	{
		rwlock_init(&(xcp_router_priv[i].lk));
		xcp_router_priv[i].input_traffic_ = 0;
		xcp_router_priv[i].avg_rtt_ = 0;
		xcp_router_priv[i].control_timeout_ = MIN_INTERVAL;
		xcp_router_priv[i].queue_timeout_ = ALLOWED_QUEUE;
		xcp_router_priv[i].sum_rtt_by_cwnd_ = 0;
		xcp_router_priv[i].sum_rttsq_by_cwnd_ = 0;
		xcp_router_priv[i].epsilonp_ = 0;
		xcp_router_priv[i].epsilonn_ = 0;
		xcp_router_priv[i].residue_pos_fbk_ = 0;
		xcp_router_priv[i].residue_neg_fbk_ = 0;
		xcp_router_priv[i].min_queue_size_ = 0x0fffffff;	
		xcp_router_priv[i].max_queue_size_ = 0;	
		init_timer(&(xcp_router_priv[i].control_timer_));
		xcp_router_priv[i].control_timer_.expires = (jiffies + (MIN_INTERVAL * HZ)/1000);
		xcp_router_priv[i].control_timer_.data = (unsigned long)&xcp_router_priv[i];
		xcp_router_priv[i].control_timer_.function = control_timeout_handler;
		add_timer(&(xcp_router_priv[i].control_timer_));
		xcp_router_priv[i].is_control_timer_up_ = 1;
		init_timer(&(xcp_router_priv[i].queue_timer_));
		xcp_router_priv[i].queue_timer_.expires = (jiffies + (ALLOWED_QUEUE * HZ)/1000);
		xcp_router_priv[i].queue_timer_.data = (unsigned long)&xcp_router_priv[i];
		xcp_router_priv[i].queue_timer_.function = queue_timeout_handler;
		add_timer(&(xcp_router_priv[i].queue_timer_));
		xcp_router_priv[i].is_queue_timer_up_ = 1;
		xcp_router_priv[i].capacity_ = CAPACITY[i];
		
	}
	printk( KERN_ALERT "#Module xcpr init\n" );
	return 0;
}

static void xcpr_exit_module(void)
{
	int i = 0;
	nf_unregister_hook(&netfilter_router_out);
	for (i = 0; i < MAX_DEVICES; i++) {
		if (xcp_router_priv[i].is_control_timer_up_==1) {
			xcp_router_priv[i].is_control_timer_up_ = 0;
			del_timer_sync(&(xcp_router_priv[i].control_timer_));
		}
		if (xcp_router_priv[i].is_queue_timer_up_==1) {
			xcp_router_priv[i].is_queue_timer_up_ = 0;
			del_timer_sync(&(xcp_router_priv[i].queue_timer_));
		}
	}
	printk( KERN_ALERT "#Module xcpr exit\n" );
}

module_init(xcpr_init_module);
module_exit(xcpr_exit_module);
